@@ -1,30 +1,56 @@
 """
-NBX-Build Instances are APIs to your machines. These APIs can be used to change state of
-the machine (start, stop, etc.), can be used to transfer files to and from the machine
-and to an extent running and managing programs (WIP).
+NBX-Build Instances are APIs to your machines. These APIs can be used to change state of the machine (start, stop,
+etc.), can be used to transfer files to and from the machine (using ``Instance.mv`` commands), calling any shell
+command using ``Instance.remote``.
+
+CLI Commands
+------------
+
+Build comes built in with several functions to move and list files from the instance. All you need to know is that
+cloud files have prefix ``nbx://`` which is where your projects are. Here's a quick list:
+
+.. code-block:: bash
+
+  # move files
+  nbx build -i 'instance_name' --workspace_id 'workspace_id' \
+    mv ./local_file nbx://cloud_file
+
+  # or move folders
+  nbx build -i 'instance_name' --workspace_id 'workspace_id' \
+    mv ./local_folder nbx://in_this/folder/
+
+
+You might be required to agree to the SSH connection being setup. If you want to avoid that set ``NBOX_SSH_NO_HOST_CHECKING=1``.
+All these APIs are also available in python.
+
+Documentation
+-------------
 """
 
-import sys
+import io
+import os
 import time
 import shlex
-from typing import List
-from functools import partial
+from json import loads
+from subprocess import run, check_output
 from tabulate import tabulate
-from tempfile import gettempdir
+from typing import Dict, List
 from requests.sessions import Session
 
-from .subway import SpecSubway,  TIMEOUT_CALLS
-from . import utils as U
-from .utils import NBOX_HOME_DIR, logger
-from .init import nbox_ws_v1, create_webserver_subway
-from .auth import secret
+import nbox.utils as U
+from nbox.auth import secret
+from nbox.utils import logger
+from nbox.subway import TIMEOUT_CALLS
+from nbox.init import nbox_ws_v1, create_webserver_subway
 
 
 ################################################################################
+'''
 # NBX-Instances Functions
-# =======================
-# The methods below are used to talk to the Webserver APIs and other methods
-# make the entire process functional.
+
+The methods below are used to talk to the Webserver APIs and other methods
+make the entire process functional.
+'''
 ################################################################################
 
 def print_status(workspace_id: str = None, fields: List[str] = None):
@@ -38,7 +64,7 @@ def print_status(workspace_id: str = None, fields: List[str] = None):
 
   fields = fields if fields != None else Instance.useful_keys
 
-  data = stub_projects()["data"]
+  data = stub_projects()
   projects = data["project_details"]
   data = [{k: projects[x][k] for k in fields} for x in projects]
   data_table = [[x[k] for k in fields] for x in data]
@@ -46,19 +72,21 @@ def print_status(workspace_id: str = None, fields: List[str] = None):
     logger.info(x)
 
 ################################################################################
-# NimbleBox.ai Instances
-# ======================
-# NBX-Instances is compute abstracted away to get the best hardware for your
-# task. To that end each Instance from the platform is a single class.
+'''
+# NimbleBox Build
+
+NBX-Instances is compute abstracted away to get the best hardware for your
+task. To that end each Instance from the platform is a single class.
+'''
 ################################################################################
 
 class Instance():
   status = staticmethod(print_status)
 
   # each instance has a lot of data against it, we need to store only a few as attributes
-  useful_keys = ["project_id", "project_name", "size_used", "size", "state",]
+  useful_keys = ["project_id", "project_name", "size_used", "size", "state"]
 
-  def __init__(self, i: str, workspace_id: int = None, cs_endpoint = "server"):
+  def __init__(self, i: str, workspace_id: int = None):
     """NBX-Build Instance class manages the both individual instance, but provides
     webserver functionality using ``nbox_ws_v1``, such as starting and stopping,
     deletion and more.
@@ -68,9 +96,9 @@ class Instance():
         workspace_id (int, optional): id of the workspace to use, if not provided
             personal workspace will be used.
         cs_endpoint (str, optional): endpoint to use for the webserver, this will connect
-          to the custom ports functionality of the instance. Defaults to ``server``, 
+          to the custom ports functionality of the instance. Defaults to ``server``,
     """
-    if i == None:
+    if not i:
       raise ValueError("Instance id must be provided, try --i='8h57f9'")
     i = str(i)
 
@@ -82,19 +110,20 @@ class Instance():
     self.size: float = None
     self.state: str = None
 
-    # create a new session for communication with the compute server, avoid using single
-    # session for conlficting headers
+    # few states are defined for use in the class
+    self.__opened = False
+
+    # create a new subway for webserver
     sess = Session()
     sess.headers.update({"Authorization": f"Bearer {secret.get('access_token')}"})
     stub_ws_instance = create_webserver_subway("v1", sess)
-
-    self.status = None # this is the status string
     if workspace_id == None:
       stub_projects = stub_ws_instance.user.projects
     else:
       stub_projects = stub_ws_instance.workspace.u(workspace_id).projects
 
-    project_details = stub_projects()["data"]["project_details"]
+    # filter and get the data
+    project_details = stub_projects()["project_details"]
     if i not in project_details:
       by_name = list(filter(lambda x: x['project_name'] == i, list(project_details.values())))
       if len(by_name) == 0:
@@ -105,31 +134,30 @@ class Instance():
     else:
       data = project_details[i]
     logger.info(f"Found instance '{data['project_name']}' ({data['project_id']})")
-
     for x in self.useful_keys:
       setattr(self, x, data[x])
-
-    self.url = secret.get('nbx_url')
-    self.cs_url = None
-    self.cs_endpoint = cs_endpoint
-
-    self.stub_ws_instance = stub_projects.u(self.project_id) # .../build/{project_id}
+    
+    # some data points require extra processing before usage
+    self.custom_ports: Dict[str, int] = loads(data["custom_ports"]) if data["custom_ports"] is not None else {}
+    self.exposed_ports: Dict[str, int] = loads(data["exposed_ports"]) if data["exposed_ports"] is not None else {}
+    self.stub_ws_instance = stub_projects.u(self.project_id)
     logger.debug(f"WS: {self.stub_ws_instance}")
-    self.__opened = False
-    self.running_scripts = []
+
+    # set values
     self.refresh()
     if self.state == "RUNNING":
-      self.start()
-    logger.debug(f"Instance: {self}")
+      self._open()
 
   def __repr__(self):
-    return f"<Instance ({', '.join([f'{k}:{getattr(self, k)}' for k in self.useful_keys + ['cs_url']])})>"
+    return f"<Instance ({', '.join([f'{k}:{getattr(self, k)}' for k in self.useful_keys])})>"
 
+  ################################################################################
   """
-  # Classmethods (WIP)
+  # Classmethods
 
   Functions to create instances.
   """
+  ################################################################################
 
   @classmethod
   def new(cls, project_name: str, workspace_id: str = None, storage_limit: int = 25, project_type = "blank") -> 'Instance':
@@ -137,17 +165,26 @@ class Instance():
       stub_all_projects = nbox_ws_v1.user.projects
     else:
       stub_all_projects = nbox_ws_v1.workspace.u(workspace_id).projects
-    out = stub_all_projects(_method = "post", project_name = project_name, storage_limit = storage_limit, project_type = project_type,
-    github_branch = "", github_link = "", template_id = 0, clone_id = 0
+    out = stub_all_projects(
+      _method = "post",
+      project_name = project_name,
+      storage_limit = storage_limit,
+      project_type = project_type,
+      github_branch = "",
+      github_link = "",
+      template_id = 0,
+      clone_id = 0
   )
-    print(out)
-    # return cls(project_name, url)
 
+    return out
+
+  ################################################################################
   """
-  # Status Methods
+  # State Methods
 
   Functions here are responsible for managing the states (metadata + turn on/off) of the instances.
   """
+  ################################################################################
 
   def is_running(self) -> bool:
     """Check if the instance is running.
@@ -159,7 +196,7 @@ class Instance():
 
   def refresh(self):
     """Update the data, get latest state"""
-    self.data = self.stub_ws_instance()["data"] # GET /user/projects/{project_id}
+    self.data = self.stub_ws_instance() # GET /user/projects/{project_id}
     for k in self.useful_keys:
       setattr(self, k, self.data[k])
 
@@ -182,7 +219,7 @@ class Instance():
       dedicated_hw = dedicated_hw,
       hw = "cpu" if gpu_count == 0 else "gpu",
       hw_config = hw_config,
-      zone = zone
+      region = zone
     )
 
     logger.info(f"Waiting for instance {self.project_name} ({self.project_id}) to start ...")
@@ -195,6 +232,19 @@ class Instance():
         raise TimeoutError("Instance did not start within timeout, please check dashboard")
     logger.info(f"Instance {self.project_name} ({self.project_id}) started")
 
+  def _open(self):
+    # now the instance is running, we can open it, opening will assign a bunch of cookies and
+    # then get us the exact location of the instance
+    if not self.__opened:
+      logger.debug(f"Opening instance {self.project_name} ({self.project_id})")
+      launch_data = self.stub_ws_instance.launch(_method = "post")
+      base_domain = launch_data['base_domain']
+      self.open_data = {
+        "url": f"{base_domain}",
+        "token": self.stub_ws_instance._session.cookies.get_dict()[f"instance_token_{base_domain}"]
+      }
+      self.__opened = True
+
   def start(
     self,
     cpu: int = 2,
@@ -202,13 +252,13 @@ class Instance():
     gpu_count: int = 0,
     auto_shutdown: int = 6,
     dedicated_hw: bool = False,
-    zone = "asia-south-1",
+    zone = "asia-east1-a",
     *,
     _ssh: bool = False
   ):
     """Start instance if not already running and loads APIs from the compute server.
 
-    Actual start is implemented in `_start` method, this combines other things 
+    Actual start is implemented in `_start` method, this combines other things
 
     Args:
         cpu (int, optional): CPU count should be one of ``[2, 4, 8]``
@@ -218,9 +268,7 @@ class Instance():
         dedicated_hw (bool, optional): If not spot/pre-emptible like machines used
         zone (str, optional): GCP cloud regions, defaults to "asia-south-1".
     """
-    if _ssh:
-      if "app.nimblebox.ai" in self.stub_ws_instance._url:
-        self.stub_ws_instance._url = self.stub_ws_instance._url.replace("app.", "app.rc.")
+    
 
     if auto_shutdown < 0:
       raise ValueError("auto_shutdown must be a positive integer (hours)")
@@ -232,68 +280,9 @@ class Instance():
       # TODO: @yashbonde: inform user in case of hardware mismatch?
       logger.info(f"Instance {self.project_name} ({self.project_id}) is already running")
 
-    # now the instance is running, we can open it, opening will assign a bunch of cookies and
-    # then get us the exact location of the instance
-    logger.debug(f"Opening instance {self.project_name} ({self.project_id})")
-    base_domain = self.stub_ws_instance.launch(_method = "post")["data"]["base_domain"]
-    secret.get("url")
-    self.open_data = {
-      "url": f"{base_domain}",
-      "token": self.stub_ws_instance._session.cookies.get_dict()[f"instance_token_{base_domain}"]
-    }
-
-    instance_url = self.open_data["url"].strip("/")
-    self.cs_url = f"{self.url}/{instance_url}"
-    if self.cs_endpoint:
-      self.cs_url += f"/{self.cs_endpoint}"
-
-    # TODO:@yashbonde remove this check, once v2 on prod
-    if self.cs_url.endswith("server"):
-      self.__opened = True
-      return
-
-    # create a speced-subway, this requires capturing the openAPI spec first
-    r = self.session.get(f"{self.cs_url}/openapi.json"); r.raise_for_status()
-    self.cs_spec = r.json()
-    self.compute_server = SpecSubway.from_openapi(self.cs_spec, self.cs_url, self.session)
-    logger.debug(f"CS: {self.compute_server}")
-
-    # now load all the functions from methods.py
-    logger.debug(f"Testing instance {self.project_name} ({self.project_id})")
-    out = self.compute_server.test()
-    with open(U.join(NBOX_HOME_DIR, "methods.py"), "w") as f:
-      f.write(out["data"])
-
-    sys.path.append(NBOX_HOME_DIR)
-    from methods import __all__ as m_all
-    for m in m_all:
-      trg_name = f"bios_{m}"
-      exec(f"from methods import {m} as {trg_name}")
-      fn = partial(
-        locals()[trg_name],
-        cs_sub = self.compute_server,
-        ws_sub = self.web_server,
-        logger = logger
-      )
-      setattr(self, m, fn)
-
-    # see if any developmental methods are saved locally
-    try:
-      from proto_methods import __all__ as p_all
-      for m in p_all:
-        trg_name = f"bios_{m}"
-        exec(f"from proto_methods import {m} as {trg_name}")
-        fn = partial(
-          locals()[trg_name],
-          cs_sub = self.compute_server,
-          ws_sub = self.web_server,
-          logger = logger
-        )
-        setattr(self, m, fn)
-    except:
-      pass
-
-    self.__opened = True
+    # prevent rate limiting
+    if not self.__opened:
+      self._open()
 
   def stop(self):
     """Stop Instance"""
@@ -328,124 +317,193 @@ class Instance():
     else:
       logger.warning("Aborted")
 
+  ################################################################################
   """
   # Interaction Methods
 
   Doing things with the project-build.
   """
+  ################################################################################
 
-  def __unopened_error(self):
+  def _unopened_error(self):
     if not self.__opened:
-      logger.error("You are trying to move files to a NOT-RUNNING instance, you will have to start the instance first:")
-      logger.error('    -         nbox.Instance(...).start(...)')
-      logger.error('    - python3 -m nbox build ... start ...')
+      logger.error(f"You are trying to move files to a {self.state} instance, you will have to start the instance first:")
+      logger.error(f'    - nbox.Instance("{self.project_id}", "{self.workspace_id}").start(...)')
+      logger.error(f'    - python3 -m nbox build --i "{self.project_id}" --workspace_id "{self.workspace_id}" start --help')
       raise ValueError("Instance is not opened, please call .open() first")
 
+  def __create_connection(self, *, port: int = 6174):
+    """This function is used to create a the Connection Manager object and not depend on `__create_threads` from ssh"""
+    from nbox.sub_utils.ssh import FileLogger, ConnectionManager
+
+    # create logging for RSock
+    folder = U.join(U.env.NBOX_HOME_DIR(), "tunnel_logs")
+    os.makedirs(folder, exist_ok=True)
+    filepath = U.join(folder, f"tunnel_{self.project_id}.log") # consistency with IDs instead of names
+    file_logger = FileLogger(filepath)
+    logger.debug(f"Logging RSock server to {filepath}")
+
+    # if "app.nimblebox.ai" in self.stub_ws_instance._url:
+    #   self.stub_ws_instance._url = self.stub_ws_instance._url.replace("app.", "app.rc.")
+    self._open() # open using the updated URL
+
+    conman = ConnectionManager(
+      file_logger = file_logger,
+      user = secret.get("username"), 
+      subdomain = self.open_data.get("url"),
+      auth = self.open_data.get("token"),
+    )
+    conman.add(localport = port, buildport = 2222, _ssh = False)
+    return conman
+
+  def __run_command(self, comm: str, port: int, return_output: bool = False) -> str:
+    connection = self.__create_connection(port = port)
+    try:
+      command = shlex.split(comm)
+      logger.info(f"Running command: {comm}")
+
+      # there are some commands that require output in which case we run the commands 
+      if return_output:
+        result = check_output(command, universal_newlines=True)
+      else:
+        result = run(command, universal_newlines=True)
+        result = result.stdout
+        result = "" if result is None else result
+    except KeyboardInterrupt:
+      logger.info("KeyboardInterrupt, closing connections")
+      result = ""
+    except Exception as e:
+      U.log_traceback()
+      result = ""
+
+    # Right now we close the connections the moment we are done with the execution, which means
+    # that RSock server is started and closed multiple times. Is this a good method, should we
+    # open a connection and keep it open till the process is done? Does that also introduce a
+    # security risk?
+    connection.quit()
+    return result.strip()
+
   def ls(self, path: str, *, port: int = 6174):
     """
     List files in a directory relative to '/home/ubuntu/project'
 
     ```
-    [nbox API] - nbox.Instance(...).ls("/)
-    [nbox CLI] - python3 -m nbox build ... ls /
+    [nbox API] - nbox.Instance(...).ls("/")
+    [nbox CLI] - python3 -m nbox build ... ls ./
     ```
     """
-    self.__unopened_error()
-
+    self._unopened_error()
+    if path.startswith("nbx://"):
+      path = path[6:]
     if path == "":
       raise ValueError("Path cannot be empty")
     path = "/home/ubuntu/project/" + path.strip("/")
-    logger.info(f"Looking in folder: {path}")
 
-    from nbox.sub_utils.ssh import _create_threads
-    connection = _create_threads(port, i = self.project_id, workspace_id = self.workspace_id, _ssh = False)
-
-    try:
-      import shlex
-      from subprocess import PIPE, run
-
-      # https://serverfault.com/questions/242176/is-there-a-way-to-do-a-remote-ls-much-like-scp-does-a-remote-copy
-      command = shlex.split(f'ssh -p {port} ubuntu@localhost ls -l {path}')
-      result = run(command, stdout=PIPE, stderr=PIPE, universal_newlines=True)
-      result = result.stdout
-    except KeyboardInterrupt:
-      logger.info("KeyboardInterrupt, closing connections")
-      result = None
-    connection.quit()
+    # RPC
+    logger.info(f"Looking at path: {path}")
+    comm = "ssh"
+    if U.env.NBOX_SSH_NO_HOST_CHECKING(False):
+      comm += " -o StrictHostKeychecking=no"
+    comm += f" -p {port} ubuntu@localhost 'ls -l {path}'"
+    result = self.__run_command(comm, port, return_output=True)
     return result
 
   def mv(self, src: str, dst: str, force: bool = False, *, port: int = 6174):
     """
     Move files to and fro NBX-Build.
-    
+
     Use 'nbx://' as prefix for Instance, all files will be placed relative to
     "/home/ubuntu/project/" folder.
     """
-    self.__unopened_error()
+    self._unopened_error()
 
     src_is_cloud = src.startswith("nbx://")
     dst_is_cloud = dst.startswith("nbx://")
 
     if src_is_cloud and dst_is_cloud:
-      # logger.error(f"Cannot move files between two instances")
       raise ValueError("Cannot move files between two instances")
     if not src_is_cloud and not dst_is_cloud:
-      # logger.error(f"Cannot move files on your machine")
       raise ValueError("Cannot move files on your machine")
-    
-    # check if this file already exists
-    cloud_file = src if src_is_cloud else dst
-    cloud_file = cloud_file.replace("nbx://", "")
-    
-    ls_res = self.ls(cloud_file)
-    # print("lllllllllll", ls_res)
-    logger.debug("ls_res", ls_res)
-    if ls_res:
-      logger.info(f"File {cloud_file} already exists")
-      if not force:
-        logger.error("Aborted, use --force to override")
-        return
-      logger.info("Overriding file")
+    elif src_is_cloud:
+      # if the dst is . or ./ it means to copy it to the folder where command is run so that
+      # is acceptable but any other path is not acceptable
+      if not dst in ["./", "."] and os.path.exists(dst) and not force:
+        raise ValueError(f"Source file '{dst}' already exists, pass --force to override")
+      ls_res = self.ls(src, port = port)
+      src_is_dir = ls_res.startswith("total")
+      if not src_is_dir:
+        logger.info(f"ls result on cloud: {ls_res}")
+    elif dst_is_cloud:
+      assert os.path.exists(src), f"Source file '{src}' does not exist"
+      # check if this file already exists on the cloud
+      ls_res = self.ls(dst, port = port)
+      logger.info(f"ls result on cloud: {ls_res}")
+      if ls_res:
+        logger.info(f"File {dst} already exists")
+        if not force:
+          logger.error("Aborted, use --force to override")
+          return
+        logger.info("Overriding file")
+      src_is_dir = os.path.isdir(src)
 
     src = "/home/ubuntu/project/" + src[6:] if src_is_cloud else src
     dst = "/home/ubuntu/project/" + dst[6:] if dst_is_cloud else dst
+
+    # RPC, note order sensitivity for src and dst along with user@host
     logger.info(f"Moving {src} to {dst}")
-    
-    # # move file
-    # from nbox.sub_utils.ssh import _create_threads
-    # connection = _create_threads(port, i = self.project_id, workspace_id = self.workspace_id)
+    comm = "scp"
+    if src_is_dir:
+      logger.debug(f"Source is a directory")
+      comm += " -r" # recursive
+    if U.env.NBOX_SSH_NO_HOST_CHECKING(False):
+      logger.debug("Host checking is disabled")
+      comm += " -o StrictHostKeychecking=no"
+    comm += f" -P {port}"
+    if src_is_cloud:
+      comm += f" ubuntu@localhost:{src} {dst}"
+    else:
+      comm += f" {src} ubuntu@localhost:{dst}"
+    result = self.__run_command(comm, port)
+    return result
 
-    # try:
-    #   import shlex
-    #   from subprocess import PIPE, run
+  def rm(self, file: str, *, port: int = 6174):
+    """
+    Remove file from NBX-Build.
+    """
+    self._unopened_error()
 
-    #   # https://serverfault.com/questions/242176/is-there-a-way-to-do-a-remote-ls-much-like-scp-does-a-remote-copy
-    #   comm = f'scp -P {port} {src} localhost:{dst}'
-    #   logger.info(comm)
-    #   command = shlex.split(comm)
-    #   result = run(command, stdout=PIPE, stderr=PIPE, universal_newlines=True)
-    # except KeyboardInterrupt:
-    #   logger.info("KeyboardInterrupt, closing connections")
-    #   result = None
-    # connection.quit()
-    # return result
+    if not file.startswith("nbx://"):
+      raise ValueError("File must be on NBX-Build, try adding 'nbx://' to start!")
+    file = "/home/ubuntu/project/" + file.replace("nbx://", "")
 
-  def rm(self):
-    pass
+    # RPC
+    logger.info(f"Removing file: {file}")
+    comm = "ssh"
+    if U.env.NBOX_SSH_NO_HOST_CHECKING(False):
+      comm += " -o StrictHostKeychecking=no"
+    comm += f" -p {port} ubuntu@localhost 'rm {file}'"
+    result = self.__run_command(comm, port)
+    return result
 
-  def run_py(self, fp: str, *args, write_fp = sys.stdout):
-    """Run any python file
-    
-    EXPERIMENTAL: FEATURES MIGHT BREAK"""
-    fp = fp.replace("nbx://", "/mnt/disks/user/project/")
-    pid = self(fp)
-    from time import sleep
-    while self(pid) == "running":
-      sleep(10)
+  def remote(self, x: str, *, port: int = 6174):
+    """
+    Run any command using SSH, in the underlying system it will setup a SSH connection and execute
+    any command.
+    """
+    self._unopened_error()
+
+    # RPC
+    logger.info(f"Running command: {x}")
+    comm = "ssh"
+    if U.env.NBOX_SSH_NO_HOST_CHECKING(False):
+      comm += " -o StrictHostKeychecking=no"
+    comm += f" -p {port} ubuntu@localhost '{x}'"
+    result = self.__run_command(comm, port)
+    return result
 
   def __call__(self, x: str):
     """EXPERIMENTAL: FEATURES MIGHT BREAK
-    
+
     Caller is the most important UI/UX. The letter ``x`` in programming is reserved the most
     arbitrary thing, and this ``nbox.Instance`` is the gateway to a cloud instance. You can:
     1. run a script on the cloud
@@ -460,47 +518,5 @@ class Instance():
     any IDE with checker is dead easy, however the performace guarantee is premium given high
     costs. Thus the logic to parsing these will have to be written as a seperate module.
     """
-    if not self.__opened:
-      raise ValueError("Instance is not opened, please call .start() first")
-    if not isinstance(x, str):
-      raise ValueError("x must be a string")
-
-      # this is to check if the PID state, returns boolean
-    if x.startswith("PID_"):
-      x = x[4:]
-      data = self.compute_server.rpc.status(x)
-      if not data["msg"] == "success":
-        raise ValueError(data["msg"])
-      
-      status = data["status"]
-      logger.debug(f"Script {x} on instance {self.project_name} ({self.project_id}) is {status}")
-
-      if status in ["stopped", "failed"]:
-        out = self.compute_server.rpc.logs(x)
-        if out["err"]:
-          logger.error(f"PID: {x} Error")
-          print(out)
-          return "error-done"
-        else:
-          return "done"
-      return status
-
-    # # assuming this is a shell command
-    # command = shlex.join(shlex.split(x)) # sanitize the input
-    comm_hash = U.hash_(x)
-    # sh_fpath = U.join(gettempdir(), "run.sh")
-    # with open(sh_fpath, "w") as f:
-    #   f.write(command)
-    # logger.info(f"Running command '{command}' [{comm_hash}] on instance {self.project_name} ({self.project_id})")
-    # nbx_fpath = f"nbx://{comm_hash}.sh"
-    # fpath = nbx_fpath.replace("nbx://", "/mnt/disks/user/project/")
-    # self.mv(sh_fpath, nbx_fpath)
-    # meta = list(filter(
-    #   lambda y: y["path"] == fpath, self.compute_server.myfiles.info(fpath)["data"]
-    # ))
-    # if not meta:
-    #   raise ValueError(f"File {x} not found on instance {self.project_name} ({self.project_id})")
-    uid = "PID_" + self.compute_server.rpc.start(x, args = ["run"])["uid"]
-    logger.info(f"Command {comm_hash} is running with PID {uid}")
-    self.running_scripts.append(uid)
-    return uid
+    raise NotImplementedError("Core talking method, not implemented yet")
+    # self._unopened_error()
