@@ -39,18 +39,51 @@ import logging
 import hashlib
 import requests
 import tempfile
+import traceback
 import randomname
 from uuid import uuid4
+from functools import partial
+from contextlib import contextmanager
+from datetime import datetime, timezone
 from pythonjsonlogger import jsonlogger
+from importlib.util import spec_from_file_location, module_from_spec
 from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed
 
+class env:
+  """
+  Single namespace for all environment variables.
+  
+  #. ``NBOX_LOG_LEVEL``: Logging level for ``nbox``, set ``NBOX_LOG_LEVEL=info|debug|warning"
+  #. ``NBOX_JSON_LOG``: Whether to print json-logs, set ``NBOX_JSON_LOG=1``
+  #. ``NBOX_JOB_FOLDER``: Folder path for the job, set ``NBOX_JOB_FOLDER=/tmp/nbox/jobs"``
+  #. ``NBOX_NO_AUTH``: If set ``secrets = None``, this will break any API request and is good only for local testing
+  #. ``NBOX_SSH_NO_HOST_CHECKING``: If set, ``ssh`` will not check for host key
+  #. ``NBOX_HOME_DIR``: By default ~/.nbx folder
+  #. ``NBOX_USER_TOKEN``: User token for NimbleBox.ai from <app.nimblebox.ai/secrets>_
+  """
+  NBOX_LOG_LEVEL = lambda x: os.getenv("NBOX_LOG_LEVEL", x)
+  NBOX_JSON_LOG = lambda x: os.getenv("NBOX_JSON_LOG", x)
+  NBOX_JOB_FOLDER = lambda x: os.getenv("NBOX_JOB_FOLDER", x)
+  NBOX_NO_AUTH = lambda x: os.getenv("NBOX_NO_AUTH", x)
+  NBOX_SSH_NO_HOST_CHECKING = lambda x: os.getenv("NBOX_SSH_NO_HOST_CHECKING", x)
+  NBOX_HOME_DIR = lambda : os.environ.get("NBOX_HOME_DIR", os.path.join(os.path.expanduser("~"), ".nbx"))
+  NBOX_USER_TOKEN = lambda x: os.getenv("NBOX_USER_TOKEN", x)
+  NBOX_NO_LOAD_GRPC = lambda: os.getenv("NBOX_NO_LOAD_GRPC", False)
+  NBOX_NO_LOAD_WS = lambda: os.getenv("NBOX_NO_LOAD_WS", False)
+
+logger = None
 
 def get_logger():
+  # add some handling so files can use functional way of getting logger
+  global logger
+  if logger is not None:
+    return logger
+
   logger = logging.getLogger("utils")
-  lvl = os.getenv("NBOX_LOG_LEVEL", "info").upper()
+  lvl = env.NBOX_LOG_LEVEL("info").upper()
   logger.setLevel(getattr(logging, lvl))
 
-  if os.environ.get("NBOX_JSON_LOG", False):
+  if env.NBOX_JSON_LOG(False):
     logHandler = logging.StreamHandler()
     logHandler.setFormatter(jsonlogger.JsonFormatter(
       '%(timestamp)s %(levelname)s %(message)s ',
@@ -69,25 +102,54 @@ def get_logger():
 
 logger = get_logger() # package wide logger
 
-def log_and_exit(msg, *args, **kwargs):
-  # convinience function to avoid tracebacks
-  logger.error(msg, *args, **kwargs)
-  sys.exit(1)
+
+@contextmanager
+def deprecation_warning(msg, remove, replace_by: str = None, help: str = None):
+  from nbox.version import __version__
+  logger.warning("Deprecation Warning")
+  logger.warning(f"  current: {__version__}")
+  logger.warning(f"  removed: {remove}")
+  logger.warning(f"      msg: {msg}")
+  if replace_by:
+    logger.warning(f"  replace: {replace_by}")
+  if help:
+    logger.warning(f"  help: {help}")
+
+
+def load_module_from_path(fn_name, file_path):
+  spec = spec_from_file_location(fn_name, file_path)
+  foo = module_from_spec(spec)
+  mod_name = get_random_name()
+  sys.modules[mod_name] = foo
+  spec.loader.exec_module(foo)
+  fn = getattr(foo, fn_name)
+  return fn
+
+
+class FileLogger:
+  """Flush logs to a file, useful when we don't want to mess with current logging"""
+  def __init__(self, filepath):
+    self.filepath = filepath
+    self.f = open(filepath, "a+")
+
+    self.debug = partial(self.log, level="debug",)
+    self.info = partial(self.log, level="info",)
+    self.warning = partial(self.log, level="warning",)
+    self.error = partial(self.log, level="error",)
+    self.critical = partial(self.log, level="critical",)
+
+  def log(self, message, level):
+    self.f.write(f"[{datetime.now(timezone.utc).isoformat()}] {level}: {message}\n")
+    self.f.flush()
+
+
+# def log_and_exit(msg, *args, **kwargs):
+#   # convinience function to avoid tracebacks
+#   logger.error(msg, *args, **kwargs)
+#   sys.exit(1)
 
 # lazy_loading/
 
-class Fn:
-  # TODO: @yashbonde build this, any arbitrary function
-  def __init__(self, fn, requirements = None):
-    self.fn = fn
-    self.requirements = requirements
-
-  def __repr__(self) -> str:
-    return f"<Fn {self.fn.__module__}.{self.fn.__qualname__}>"
-
-  def __call__(self, *args, **kwargs):
-    return self.fn(*args, **kwargs)
-
 def isthere(*packages, soft = True):
   """Checks all the packages
 
@@ -95,7 +157,7 @@ def isthere(*packages, soft = True):
       soft (bool, optional): If ``False`` raises ``ImportError``. Defaults to True.
   """
   def wrapper(fn):
-    _fn_ = Fn(fn, packages)
+    _fn_ = fn
     def _fn(*args, **kwargs):
       # since we are lazy evaluating this thing, we are checking when the function
       # is actually called. This allows checks not to happen during __init__.
@@ -132,14 +194,15 @@ def isthere(*packages, soft = True):
 
 # path/
 
-def get_files_in_folder(folder, ext = ["*"]):
+def get_files_in_folder(folder, ext = ["*"], abs_path: bool = True):
   """Get files with ``ext`` in ``folder``"""
   # this method is faster than glob
   import os
   all_paths = []
   _all = "*" in ext # wildcard means everything so speed up
 
-  for root,_,files in os.walk(folder):
+  folder_abs = os.path.abspath(folder) if abs_path else folder
+  for root,_,files in os.walk(folder_abs):
     if _all:
       all_paths.extend([join(root, f) for f in files])
       continue
@@ -179,7 +242,6 @@ def from_pickle(path):
   with open(path, "rb") as f:
     return dill.load(f)
 
-NBOX_HOME_DIR = os.environ.get("NBOX_HOME_DIR", join(os.path.expanduser("~"), ".nbx"))
 
 # /path
 
@@ -195,6 +257,14 @@ def hash_(item, fn="md5"):
   """Hash sting of any item"""
   return getattr(hashlib, fn)(str(item).encode("utf-8")).hexdigest()
 
+
+def log_traceback():
+  # f = io.StringIO("")
+  # traceback.print_exception(*sys.exc_info(), file = f)
+  f = traceback.format_exc()
+  for _l in f.splitlines():
+    logger.error(_l.rstrip())
+
 # /misc
 
 # datastore/
